@@ -1,5 +1,6 @@
 use crate::IConfiguration;
 use serde_json::json;
+use std::fmt::{self, Formatter};
 use std::str::FromStr;
 use std::time::Duration;
 
@@ -16,7 +17,6 @@ pub use oauth2::{AccessToken, RefreshToken};
 use oauth2::{
     AuthUrl, AuthorizationCode, ClientId, CsrfToken, PkceCodeChallenge, PkceCodeVerifier, TokenUrl,
 };
-use serde::{de::Visitor, Deserialize, Deserializer, Serialize, Serializer};
 
 type Client = oauth2::Client<
     BasicErrorResponse,
@@ -27,12 +27,92 @@ type Client = oauth2::Client<
     BasicRevocationErrorResponse,
 >;
 
+// FIXME: Implement errors around the IndieAuth flow.
+#[derive(thiserror::Error, Debug)]
+pub enum Error {
+    #[error("An IndieAuth operation failed to completeE: {0}")]
+    Other(anyhow::Error),
+}
+
 // This module holds abstractions for interacting with IndieAuth servers,
 // effectively making this a 'client' module.
 
-#[derive(Clone, Debug, Default)]
-pub struct Scopes {
-    scopes: Vec<Scope>,
+#[derive(Clone, Debug, Default, PartialEq)]
+pub struct Scopes(Vec<Scope>);
+
+impl ToString for Scopes {
+    fn to_string(&self) -> String {
+        self.0
+            .iter()
+            .map(|s| s.to_string())
+            .collect::<Vec<String>>()
+            .join(" ")
+    }
+}
+
+impl FromStr for Scopes {
+    type Err = std::convert::Infallible;
+
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        Ok(Self(
+            s.split(" ")
+                .map(|st| Scope::new(st.to_string()))
+                .collect::<Vec<Scope>>(),
+        ))
+    }
+}
+
+impl Scopes {
+    pub fn is_empty(&self) -> bool {
+        self.0.is_empty()
+    }
+
+    pub fn has(&self, scope: &str) -> bool {
+        self.0.iter().any(|s| s.to_string().starts_with(scope))
+    }
+
+    pub fn as_vec(&self) -> &Vec<Scope> {
+        &self.0
+    }
+}
+
+impl serde::Serialize for Scopes {
+    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+    where
+        S: serde::Serializer,
+    {
+        serializer.serialize_str(&self.to_string())
+    }
+}
+
+struct ScopesVisitor;
+
+impl<'de> serde::de::Visitor<'de> for ScopesVisitor {
+    type Value = Scopes;
+
+    fn expecting(&self, formatter: &mut Formatter) -> fmt::Result {
+        formatter.write_str("a list of strings or a string of space-separated values")
+    }
+
+    fn visit_str<E>(self, s: &str) -> Result<Self::Value, E>
+    where
+        E: serde::de::Error,
+    {
+        Ok(Scopes(
+            s.split(" ")
+                .map(|scope_str| Scope::new(scope_str.to_string()))
+                .collect::<Vec<_>>(),
+        ))
+    }
+}
+
+impl<'de> serde::Deserialize<'de> for Scopes {
+    fn deserialize<D>(d: D) -> Result<Self, D::Error>
+    where
+        D: serde::Deserializer<'de>,
+    {
+        d.deserialize_str(ScopesVisitor)
+    }
 }
 
 #[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
@@ -40,7 +120,13 @@ pub struct RedeemPayload {
     pub code: String,
     pub client_id: url::Url,
     pub redirect_uri: url::Url,
-    pub code_verifier: String,
+    pub code_verifier: Option<String>,
+}
+
+#[derive(serde::Deserialize, serde::Serialize, Debug, Clone)]
+#[serde(tag = "grant_type", rename_all = "snake_case")]
+pub enum GrantPayload {
+    AuthorizationCode(RedeemPayload),
 }
 
 #[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
@@ -57,87 +143,18 @@ pub struct CodeResponse {
     profile: Option<Profile>,
 }
 
-struct ScopesVisitor {}
-
-impl<'de> Visitor<'de> for ScopesVisitor {
-    type Value = Scopes;
-
-    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
-        formatter.write_str("a string of words")
-    }
-
-    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E> {
-        Ok(Scopes::from_str(v).unwrap())
-    }
-}
-
-impl<'de> Deserialize<'de> for Scopes {
-    fn deserialize<D>(deserializer: D) -> Result<Scopes, D::Error>
-    where
-        D: Deserializer<'de>,
-    {
-        deserializer.deserialize_str(ScopesVisitor {})
-    }
-}
-
-impl Serialize for Scopes {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
-    where
-        S: Serializer,
-    {
-        serializer.serialize_str(&self.to_string())
-    }
-}
-
-impl ToString for Scopes {
-    fn to_string(&self) -> String {
-        self.scopes
-            .iter()
-            .map(|s| s.to_string())
-            .collect::<Vec<String>>()
-            .join(" ")
-    }
-}
-
-impl FromStr for Scopes {
-    type Err = std::convert::Infallible;
-
-    fn from_str(s: &str) -> Result<Self, Self::Err> {
-        Ok(Self {
-            scopes: s
-                .split(" ")
-                .map(|st| Scope::new(st.to_string()))
-                .collect::<Vec<Scope>>(),
-        })
-    }
-}
-
-impl Scopes {
-    pub fn is_empty(&self) -> bool {
-        self.scopes.is_empty()
-    }
-
-    pub fn has(&self, scope: &str) -> bool {
-        self.scopes.iter().any(|s| s.to_string().starts_with(scope))
-    }
-
-    pub fn to_vec(&self) -> Vec<Scope> {
-        self.scopes.clone()
-    }
-}
-
 #[derive(Default, Debug)]
 pub struct DiscoveryResult {
     authorization_endpoints: Vec<String>,
     token_endpoints: Vec<String>,
 }
 
-/// Provides the basis of making a IndieAuth request.
 pub struct AuthenticationRequest {
     client: Box<Client>,
     me: String,
     scopes: Vec<String>,
     verifier: PkceCodeVerifier,
+    challenge: PkceCodeChallenge,
     redirect_uri: Option<String>,
     client_id: String,
 }
@@ -159,6 +176,7 @@ impl AuthenticationRequest {
         self.verifier.secret().clone()
     }
 
+    /// Creates a new request to obtain authentication information from a IndieAuth server.
     pub fn new(
         me: &str,
         client_id: &str,
@@ -169,10 +187,11 @@ impl AuthenticationRequest {
         let client = Box::new(Client::new(
             ClientId::new(client_id.to_string()),
             None,
-            AuthUrl::new(authorization_endpoint.to_string()).map_err(|e| anyhow!("{:#?}", e))?,
+            AuthUrl::new(authorization_endpoint.to_string())
+                .map_err(|e| anyhow!("Failed to parse the authorization endpoint URL: {:#?}", e))?,
             None,
         ));
-        let verifier = PkceCodeChallenge::new_random_sha256().1;
+        let (challenge, verifier) = PkceCodeChallenge::new_random_sha256_len(48);
 
         Ok(Self {
             client,
@@ -180,6 +199,7 @@ impl AuthenticationRequest {
             me: me.to_string(),
             scopes: scopes.unwrap_or_default(),
             verifier,
+            challenge,
             redirect_uri,
         })
     }
@@ -195,12 +215,13 @@ impl AuthenticationRequest {
         let redirect_url = RedirectUrl::from_url(url::Url::from_str(&redirect_uri)?);
         let request = self
             .client
-            .authorize_url(CsrfToken::new_random)
-            .set_pkce_challenge(PkceCodeChallenge::from_code_verifier_sha256(&self.verifier))
+            .authorize_url(|| CsrfToken::new_random_len(48))
+            .set_pkce_challenge(self.challenge.clone())
             .add_extra_param("me", self.me.clone())
             .add_scope(Scope::new(self.scopes.join(" ")))
             .set_redirect_uri(std::borrow::Cow::Borrowed(&redirect_url));
-        let (url, state) = request.url();
+        let (url, state) = dbg!(request.url());
+
         Ok((url.to_string(), state.secret().clone()))
     }
 }
@@ -229,18 +250,21 @@ pub struct Token {
 
 impl TokenResponse<BasicTokenType> for Token {
     fn scopes(&self) -> Option<&Vec<Scope>> {
-        // FIXME: This knows too much - is there a better way to do this?
-        Some(&self.scopes.scopes)
+        Some(self.scopes.as_vec())
     }
+
     fn access_token(&self) -> &AccessToken {
         &self.access_token
     }
+
     fn refresh_token(&self) -> Option<&RefreshToken> {
         self.refresh_token.as_ref()
     }
+
     fn expires_in(&self) -> Option<Duration> {
         self.expires_in.map(|s| Duration::from_secs(s))
     }
+
     fn token_type(&self) -> &BasicTokenType {
         &BasicTokenType::Bearer
     }
@@ -277,7 +301,7 @@ where
 {
     let resp = cfg
         .request(Some("Requesting information about a token".to_owned()))
-        .request(reqwest::Method::POST, endpoint.to_owned())
+        .request(reqwest::Method::GET, endpoint.to_owned())
         .header("accept", "application/json")
         .bearer_auth(access_token)
         .send()
@@ -305,7 +329,7 @@ pub async fn redeem<'a, C, R>(
 ) -> Result<R, crate::Error>
 where
     C: IConfiguration + Send + Sync,
-    R: for<'de> Deserialize<'de>,
+    R: for<'de> serde::Deserialize<'de>,
 {
     let endpoint_resp = cfg
         .request(Some(
@@ -348,15 +372,16 @@ mod tests {
 
     #[test]
     fn AuthenticationRequest_construct_url_has_expected_params() {
-        let me = "https://jacky.wtf";
-        let client_id = "http://sele.black.af";
-        let endpoint = "https://sele.black.af/endpoints/wow";
-        let redirect_uri = "https://special-domain-name.endpoints.sele.black.af/testing_page";
+        let me = "https://jacky.wtf".to_string();
+        let client_id = "http://sele.black.af".to_string();
+        let endpoint = "https://sele.black.af/endpoints/wow".to_string();
+        let redirect_uri =
+            "https://special-domain-name.endpoints.sele.black.af/testing_page".to_string();
         let result = AuthenticationRequest::new(
-            me,
-            client_id,
-            endpoint,
-            Some(redirect_uri.to_owned()),
+            &me,
+            &client_id,
+            &endpoint,
+            Some(redirect_uri.clone()),
             None,
         );
 
@@ -375,40 +400,31 @@ mod tests {
         ));
         let challenge_str = challenge.as_str();
 
-        assert!(dbg!(url.clone()).contains(
-            serde_urlencoded::to_string(&[("redirect_uri", &redirect_uri)])
-                .unwrap()
-                .as_str()
-        ));
-        assert!(url.contains(
-            serde_urlencoded::to_string(&[
-                ("code_challenge", challenge_str),
-                ("code_challenge_method", "S256"),
-            ])
-            .unwrap()
-            .as_str()
-        ));
+        use std::collections::HashMap;
 
-        assert!(url.contains(
-            serde_urlencoded::to_string(&[("client_id", &client_id)])
-                .unwrap()
-                .as_str()
-        ));
-        assert!(url.contains(
-            serde_urlencoded::to_string(&[("me", &me)])
-                .unwrap()
-                .as_str()
-        ));
-        assert!(url.contains(
-            serde_urlencoded::to_string(&[("state", &state)])
-                .unwrap()
-                .as_str()
-        ));
-        assert!(url.contains(
-            dbg!(serde_urlencoded::to_string(&[("scope", "read create")]))
-                .unwrap()
-                .as_str()
-        ));
+        let params: HashMap<String, String> =
+            dbg!(serde_qs::from_str(url.parse::<url::Url>().unwrap().query().unwrap()).unwrap());
+        assert!(params.contains_key(&"redirect_uri".to_string()));
+        assert_eq!(
+            params.get("scope").cloned(),
+            Some("read create".to_string())
+        );
+        assert_eq!(
+            params.get("code_challenge").cloned(),
+            Some(challenge_str.to_string())
+        );
+        assert_eq!(
+            params.get("response_type").cloned(),
+            Some("code".to_string())
+        );
+        assert_eq!(
+            params.get("code_challenge_method").cloned(),
+            Some("S256".to_string())
+        );
+        assert_eq!(params.get("state").cloned(), Some(state));
+        assert_eq!(params.get("me").cloned(), Some(me));
+        assert_eq!(params.get("client_id").cloned(), Some(client_id));
+        assert_eq!(params.get("redirect_uri").cloned(), Some(redirect_uri));
     }
 
     #[test]
@@ -467,7 +483,7 @@ mod tests {
             "scope": "read profile email"
         })
         .to_string();
-        let endpoint_mock = mockito::mock("POST", "/endpoint")
+        let endpoint_mock = mockito::mock("GET", "/endpoint")
             .with_body(&body)
             .expect_at_most(1)
             .create();
@@ -507,7 +523,7 @@ mod tests {
             .create();
         let payload = RedeemPayload {
             code: "a-code".to_owned(),
-            code_verifier: "a-code-verifier".to_owned(),
+            code_verifier: Some("a-code-verifier".to_owned()),
             client_id: "https://indieweb.org".parse().unwrap(),
             redirect_uri: "https://indieweb.org/back".parse().unwrap(),
         };
