@@ -12,5 +12,11 @@ impl Configuration {
 }
 
 pub fn init() -> Configuration {
+    env_logger::builder()
+        .filter_level(log::LevelFilter::max())
+        .is_test(true)
+        .try_init()
+        .ok();
+    log::debug!("Ready");
     Configuration::new()
 }
